#lang racket

(provide interp)

; e ::= (letrec ([x (lambda (x ...) e)]) e)
;     | (let ([x e] ...) e)
;     | (lambda (x ...) e)
;     | (e e ...)
;     | x
;     | (and e e) | (or e e)
;     | (if e e e)
;     | (uprim e) | (bprim e e)
;     | datum
; datum ::= nat | (quote ()) | #t | #f
; nat ::= 0 | 1 | 2 | ...
; x is a symbol
; uprim ::= add1 | car | cdr | null? | not
; bprim ::= + | * | cons

; to ====>

; e ::= (lambda (x) e)
;    | (e e)
;    | x

(define (interp e)
  (define (prim? p)
    (if (member p '(+ * cons add1 car cdr null? not)) #t #f))
  
  (define (ext-env env xs vs)
          (foldl (lambda (x v env) (hash-set env x v)) env xs vs))
  
  (define (interp-env e env)
    (match e
      [`(letrec ([,f (lambda (,xs ...) ,e0)]) ,e1)
       (letrec ([fv (lambda vs (interp-env e0 (ext-env (hash-set env f fv) xs vs)))])
         (interp-env e1 (hash-set env f fv)))]
      
      [`(let ([,xs ,e0s] ...) ,e1)
       (interp-env e1 (ext-env env xs (map (lambda (e) (interp-env e env)) e0s)))]
      
      [(? symbol? x)
       (hash-ref env x)]
      
      [`(lambda (,xs ...) ,e)
       (lambda vs (interp-env e (ext-env env xs vs)))]

      ; Except in assignment 1, only booleans will be used with and, or, if forms
      [`(and ,e0 ,e1)
       (define v0 (interp-env e0 env))
       (if v0
           (interp-env e1 env)
           #f)]
      
      [`(or ,e0 ,e1)
       (define v0 (interp-env e0 env))
       (if v0
           v0
           (interp-env e1 env))]
      
      [`(if ,e0 ,e1 ,e2)
       (if (interp-env e0 env)
           (interp-env e1 env)
           (interp-env e2 env))]
      
      [(? boolean? b) b]
      [(? number? n) n]
      [''() '()]

      ; We use the eval to turn 'cons => cons
      [`(,(? prim? p) ,es ...)
       (apply (eval p (make-base-namespace))
              (map (lambda (e) (interp-env e env)) es))]

      ; This must come last because all other patterns guard it
      [`(,efun ,es ...)
       (define vfun (interp-env efun env))
       (define vs (map (lambda (e) (interp-env e env)) es))
       (apply vfun vs)]))

  (interp-env e (hash)))







